program mm_idrs

   use matrix_market
   use general_module
   use idrs_module
   use plot_ascii

   implicit none 

! Input parameters that define the problem:
   character*20                  :: matrix_name, default_matrix_name

! IDRS parameters:
   integer                       :: maxit, default_maxit
   integer                       :: s, default_s
   real(kind=rp)                 :: tol, default_tol
   real(kind=rp)                 :: relres
   character*8                   :: variant
   integer                       :: method
   integer                       :: flag              
   integer                       :: iter
  
! Matrices, number of equations etc.:
   integer                       :: neq, nrhs, ierr
   type(matrix)                  :: A, RHS
   type(preconditioner)          :: M1
   character*30                  :: matrix_file, rhs_file
   real(kind=rp), allocatable    :: xr(:,:), br(:,:)
   complex(kind=cp), allocatable :: xc(:,:), bc(:,:)

   integer                       :: t0, t1, clock_rate, clock_max

! Parameter for plotting:
   logical                       :: plot, default_plot
   logical                       :: ocean

! Is the rhs-vector given:
   logical                       :: rhs_given

! Initialization for parameter routines (read command line)
   call initialize( )

! Read the parameters
   default_matrix_name = ''
   matrix_name = get_parameter( '-mm', default_matrix_name )
   if ( matrix_name == '' ) then
      write(*,*)
      write(*,*) 'idrs allows you to test different IDR(s) versions on matrix-market test problems.'
      write(*,*) 'The programme without command-line options gives this short help.'
      write(*,*) 'The following options are available:'
      write(*,*) '   -mm [matrix name] ! Matrix name should be given without .mtx extension.'
      write(*,*) '                     ! If the file [matrix name]_b exists, it is taken as RHS, else 1 is taken as RHS.'
      write(*,*) '   -biortho [s]      ! Use the biortho variant with parameter s.'
      write(*,*) '   -minsync [s]      ! Use the minsync variant with parameter s.'
      write(*,*) '                     ! If no variant is specified, standard bicgstab is used.'
      write(*,*) '   -maxit [maxit]    ! Specifies maximum number of iterations (default:4000).'
      write(*,*) '   -tol [tol]        ! Specifies tolerance (default:1e-6).'
      write(*,*) '   -plot             ! Makes ascii-plot of ocean examples (stommel and sag matrices).' 
      write(*,*)
      write(*,*) 'Test problems in matrix-market format should be stored in the directory MATRICES.'
      write(*,*) 'If a file [matrix name]_b.mtx is provided it is taken as RHS.'
      write(*,*) 'If not, a constant RHS-vector is taken consisting of ones.'
      write(*,*) 'The directory MATRICES contains test problems with the names stommel?.mtx and sag?.mtx.'
      write(*,*) 'These are all global ocean circulation problems. The numbers in the names correspond to the grid spacing.'
      write(*,*) 'The files stommel?_b.mtx and sag?_b.mtx each contain 12 RHS-vectors for wind-field data.'
      write(*,*) 'The directory MATRICES also contains test problems with the name wedge?.mtx.'
      write(*,*) 'These are (complex) Helmholtz problems. The numbers in the name refer to the grid refinement level.'
      write(*,*) 'Additional test problems can be downloaded for example from the SuiteSparse Matrix Collection.'
      write(*,*) 
      write(*,*) 'IDR(s) is used with a Jacobi preconditioner. '
      write(*,*) 'If the main diagonal contains zeros, no preconditioner is used.'
      write(*,*) 
      write(*,*) 'Example: idrs -mm stommel4 -biortho 4 -maxit 1000 -tol 1e-8 -plot'
      write(*,*)
   else

      default_maxit = 4000
      maxit = get_parameter('-maxit', default_maxit )
      default_tol = 1e-6
      tol = get_parameter('-tol', default_tol )
      default_plot = .false.
      plot = get_parameter('-plot', default_plot )
      default_s = 0
      s = get_parameter('-biortho', default_s )    
      if ( s > 0 ) then
         variant = 'biortho'
      else
         s = get_parameter('-minsync', default_s )    
         if ( s > 0 ) then
            variant = 'minsync'
         else
            variant = 'bicgstab'
            s = 1
         end if
      end if

! Names of the files with the matrix and right-hand-side vector:
      write(matrix_file,'(a,a,a)') 'MATRICES/',trim(matrix_name),'.mtx'
      write(rhs_file,'(a,a,a)') 'MATRICES/',trim(matrix_name),'_b.mtx'

! Read the system matrix:
      open(unit = 10, file = matrix_file, status = 'old', iostat = ierr )
      if ( ierr > 0 ) stop 'Error while opening matrix file '
      call read_mm_matrix( 10, A )
      close(10)
      if ( A%rows .ne. A%cols ) stop 'Matrix is not square'
      neq  = A%rows
 
! Construct the (diagonal) preconditioner:
      call make_preconditioner( A, M1 )

! Read the right-hand-side matrix:
      open(unit = 20, file = rhs_file, status = 'old', iostat = ierr )
      if ( ierr > 0 ) then
         rhs_given = .false.
         nrhs = 1
      else
         rhs_given = .true.
         call read_mm_matrix( 20, RHS )
         close(20)
         nrhs = RHS%cols
      end if

! Allocate space for the solution vectors and rhs-vector
      if ( A%field == 'complex' ) then
         allocate( xc(neq,nrhs), bc(neq,nrhs) )
         bc = (1.,0)
      else
         allocate( xr(neq,nrhs), br(neq,nrhs) )
         br = 1.
      end if

      if ( rhs_given ) then 
! Copy the rhs-vector b from the user-defined type RHS
         if ( A%field == 'complex' .and. RHS%field == 'complex' ) then
            bc = reshape( RHS%cval, (/neq, nrhs/) )
         elseif ( A%field == 'complex' .and. RHS%field == 'real' ) then
            bc = reshape( RHS%rval, (/neq, nrhs/) )
         elseif ( A%field == 'real' .and. RHS%field == 'complex' ) then
            stop 'Complex RHS not allowed for real matrix '
         elseif ( A%field == 'real' .and. RHS%field == 'complex' ) then
            br = reshape( RHS%rval, (/neq, nrhs/) )
         end if
      end if

! Solve the system
      call system_clock ( t0, clock_rate, clock_max )
      if ( A%field == 'complex' ) then
         xc = idrs( A, bc, M1, s, tol, maxit, variant, flag, relres, iter )
      else
         xr = idrs( A, br, M1, s, tol, maxit, variant, flag, relres, iter )
      end if
      call system_clock ( t1, clock_rate, clock_max )

! Output: 
      write(*,*)
      write(*,'(a)') '======================================================='
      write(*,'(a,a)')      'Matrix                         = ', matrix_name
      if ( rhs_given ) then
         write(*,'(a)')     'RHS vector                     = given'
      else
         write(*,'(a)')     'RHS vector                     = 1'
      end if
      write(*,'(a,a)')      'The preconditioner is          = ', M1%which_preconditioner 
      write(*,'(a,i5)')     'The number of equations        = ', neq
      write(*,'(a,i5)')     'The number of right-hand-sides = ', nrhs
      write(*,'(a,a)')      'The type of the matrix         = ', A%field
      write(*,'(a,a)')      'The structure of the matrix    = ', A%symm
      write(*,*)
      if ( variant == 'bicgstab' ) then
         write(*,'(a,a,a,i4)') 'Results for ',variant
      else
         write(*,'(a,a,a,i4)') 'Results for ',variant,' with s    =  ', s
      end if
      write(*,'(a,f7.3,a)') 'Elapsed time                   = ', real(t1-t0)/real(clock_rate), 's'
      write(*,'(a,i5)')     'Number of iterations           = ', iter
      write(*,'(a,e7.2)')   'Relative residual norm         = ', relres
      if ( flag > 0 ) then
         if ( flag == 1 ) write(*,'(a)') 'Maximum number of iterations reached!'
         if ( flag == 2 ) write(*,'(a)') 'Accuracy above prescribed tolerance!'
         if ( flag == 3 ) write(*,'(a)') 'Break down!'
      end if
      write(*,'(a)') '======================================================='

! Plot the results (only plots ocean models)

      ocean = ( index(matrix_name,'sag') > 0  .or. index(matrix_name,'stommel') > 0 )
      if ( plot .and. ocean ) call plot_ocean( xr, neq, nrhs, matrix_name )
   end if

end program mm_idrs

subroutine plot_ocean( x, neq, nrhs, matrix_name )

!
! Plot routine for the Stommel and Sag ocean problems
!

   use matrix_market
   use plot_ascii

   implicit none

   character*20               :: matrix_name, model

   type(matrix)               :: P
   integer                    :: grid
   real(kind=rp)              :: x(neq,nrhs)
   real(kind=rp), allocatable :: SOL(:,:)
   real,          allocatable :: f(:,:)
   character*30               :: map_file

   integer                    :: ndof, mdof, npx, npy, neq, irhs, nrhs, i
   real                       :: cmin, cmax
   character*80               :: month(12)
  
   month = (/'January   ','February  ','March     ','April     ','May       ',&
             'June      ','July      ','August    ','September ','October   ',&
             'November  ','December  '/) 

! First determine which model (Sag or Stommel) and wihich grid
   i = index( matrix_name, 'sag' )
   if ( i > 0 ) then
! Sag model
      read(matrix_name,'(a3,i1)') model, grid
   else
      i = index( matrix_name, 'stommel' )
      if ( i > 0 ) then
! Stommel model
         read(matrix_name,'(a7,i1)') model, grid
      else
! Not an ocean model, return
         return
      end if
   end if

! Read the mapping file (to earth dofs)
   write(map_file,'(a,i1,a)') 'MATRICES/grid',grid,'_map.mtx'
   open(unit = 30, file = map_file )
   call read_mm_matrix( 30, P )
   close(30)
   mdof = P%rows
   ndof = P%cols
   allocate( SOL(mdof,nrhs) )

! Map solution to earth-coordinates
   SOL = P*x

! How many pixels in each direction?
   npx = 360/grid+1
   npy = 180/grid-1

! Now plot the solutions
   allocate( f(npx,npy) )
   cmin = real(minval( SOL ))
   cmax = real(maxval( SOL ))
   do irhs = 1,nrhs
      call csc
      f = reshape( real(SOL(:,irhs)), (/npx, npy/) )
      call image( f, month(irhs), 'greysc', cmin, cmax )
      call plot_continents( P, npx, npy )
      read(*,*)
   end do

end subroutine plot_ocean

subroutine plot_continents( P, nx, ny )

   use matrix_market
   use plot_ascii

   integer                        :: ix, iy, nx, ny, m, n, i, k, mx, ncont
   character(len=12), allocatable :: colour(:)
   type(matrix)                   :: P
   integer, allocatable           :: ppv(:)      ! Pixels per variable 
   integer                        :: cont(1)

   colour = palet( 'normal' )
   ncol   = size(colour)
   ncont = ncol

   m = P%rows
   n = P%cols

   allocate(ppv(n))
   ppv = 0
   
! Determine the largest continents:
   do i = 1, m 
      ppv(P%jndx(i)) = ppv(P%jndx(i)) + 1
   end do

   do k = 1, ncont
      cont = maxloc( ppv ) 
      mx = ppv(cont(1))
      if ( mx == 2 ) exit
      ppv(cont(1)) = 0
! Plot this continent:
      do i = 1, m
         if ( P%jndx(i) .eq. cont(1) ) then 
            ix = mod(P%indx(i)-1,nx)+1
            iy = (P%indx(i)-1)/nx+1
            call pixel(colour(k),ix,ny-iy+1)
         end if
      end do
   end do

   call cup(1,ny+2)
   call sgr

end subroutine plot_continents
